/* yaf-splash - yet another fine splash program for x
 * Copyright (c) 2002, 2003 John Morton 
 *
 * Contains gobs of code wrenched wholesale from xdaliclock and xsplash,
 * particularly the word iterator and trasparency stuff.
 * Copyright (c) 2000, 2002 Jamie Zawinski <jwz@jwz.org>
 *
 * While the outline code was lifted from xsublim
 * Copyright (c) 1999 - 2000 Greg Knauss (greg@eod.com) 
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation.  No representations are made about the suitability of this
 * software for any purpose.  It is provided "as is" without express or 
 * implied warranty.
 *
 * yaf-splash --- pops up a window of text, auto-sized and centered on the screen,
 * that dismisses when you click on it, rather like xsplash, from Gronk. It can 
 * also display the text with a transparent background via shaped windows, as well
 * as displaying a clock instead of text.
 */

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <time.h>

#include <X11/Xlib.h>
#include <X11/Xresource.h>
#include <X11/Xos.h>
#include <X11/Intrinsic.h>
#include <X11/IntrinsicP.h>
#include <X11/CoreP.h>
#include <X11/Shell.h>
#include <X11/StringDefs.h>
#include <X11/Xatom.h>
#include <X11/keysym.h>
#include <X11/Xmu/Error.h>

#ifdef HAVE_SHAPE
# include <X11/extensions/shape.h>
#endif

#include "yaf-splash.h"
#include "version.h"
#include "resources.h"
#include "visual.h"
#include "usleep.h"
#include "vroot.h"

#define clock_usleep screenhack_usleep
#define TICK_SLEEP()    clock_usleep (80000L)           /* 8/100th second */

#undef countof
#define countof(x) (sizeof((x))/sizeof(*(x)))

#undef MAX
#undef MIN
#define MAX(a, b) ((a)>(b)?(a):(b))
#define MIN(a, b) ((a)<(b)?(a):(b))

#define CHAR_W(ch) (font->per_char \
                     ? font->per_char[(ch) - font->min_char_or_byte2].width \
                     : font->min_bounds.width)

char *progclass = "XSMsg";
char *progname = 0;
XrmDatabase db;

static Atom XA_WM_PROTOCOLS, XA_WM_DELETE_WINDOW;

typedef struct {
  unsigned long fg, bg, bd;
  int margin, border, outline; 
  int max_x, max_y;  /* The size of the text area (origin 0,0) */
  int width, height; /* The size of the window/pixmap. This is the 
			either the text area, plus any margin, or
			something supplied from the geometry resource */
  int xoff, yoff;    /* The offset origin of the text area inside the 
			window. We center this to whatever size the
			window ends up, as best we can. */
  int x,y;           /* x and y of the window */
  int changed_p;     /* Has the message changed recently? */
  int mirror_p;      /* Produce the mirror image of the text, for back projection */
  int do_shape;      /* Use the SHAPE extension to get transparent messages */
  int do_overlay;    /* Use overlays to get transparent messages */
  Display *display;
  Window window;
  Window parent;
  GC window_draw_gc;
  GC window_erase_gc;
  GC outline_draw_gc;  /* This GC fills in just the body of the outline with the
			 background colour, as the outline is in the foreground color.
		      */
  Pixmap text_pix;       /* The quantity of pixmaps are getting a little out of */
  Pixmap m_text_pix;     /* control. Might need to try a different draw strategy */
  Pixmap outline_pix;    /* than using planes? */
  Pixmap m_outline_pix;
  GC pixmap_draw_gc;
  GC pixmap_erase_gc;
  XFontStruct *font;
} draw_struct;

typedef struct {

  int do_clock;
  int do_seconds;
  int clock_mode;
  time_t countdown;
  time_t last_time;
  int countdown_quit_p; /* Whether to exit when the coundown reaches 0, or
			   continue counting back up */
} clock_struct;


char *defaults[] = {
  "*background:	   White",
  "*foreground:	   Black",
  "*borderWidth:   2",
  "*borderColor:   Black",
  "*outlineWidth:  1",
  "*margin:	   0",
  "*text:	   Obey.",
  "*font:	   -*-avantgarde-demi-r-*-*-*-720-*-*-p-*-*-*",
  "*timeout:	   0",
  "*useclock:      False",
  "*clockmode:     12",
  "*countdownquit: True",
  "*seconds:       True",
  "*transparent:   False",
  "*mirror:        False",
  /* "*geometry: +0+0", */ /* no default for this */
  /* "*placement:   Center", */
  0
};

XrmOptionDescRec options [] = {
  { "-outline",         ".outlineWidth",  XrmoptionSepArg, 0 },  
  { "-margin",          ".margin",        XrmoptionSepArg, 0 },
  { "-text",		".text",	  XrmoptionSepArg, 0 },
  { "-font",		".font",	  XrmoptionSepArg, 0 },
  { "-timeout",		".timeout",	  XrmoptionSepArg, 0 },
  { "-placement",       ".placement",     XrmoptionSepArg, 0 },
  { "-clock",           ".useclock",      XrmoptionNoArg, "True" },
  { "-12",              ".clockmode",     XrmoptionNoArg, "12" },
  { "-24",              ".clockmode",     XrmoptionNoArg, "24" },
  { "-seconds",         ".seconds",       XrmoptionNoArg, "True" },
  { "-noseconds",       ".seconds",       XrmoptionNoArg, "False" },
  { "-countdown",       ".countdown",     XrmoptionSepArg, 0 },
  { "-countdownquit",   ".countdownquit", XrmoptionNoArg, "True" },
  { "-nocountdownquit", ".countdownquit", XrmoptionNoArg, "False" },
  { "-transparent",     ".transparent",   XrmoptionNoArg, "True" },
  { "-nontransparent",  ".transparent",   XrmoptionNoArg, "False" },
  { "-mirror",          ".mirror",        XrmoptionNoArg, "True" },
  { "-nomirror",        ".mirror",        XrmoptionNoArg, "False" }
};


/* FIXME - I should have function declarations for all the functions. 
   I probably should have them in the .h file, but dammit Jim, I'm a 
   sysadmin, not a C programmer!
*/
static void iterate_words (draw_struct *state, const unsigned char *string,	       
			   int max_width, 
			   void (*mapper) (draw_struct *state, const char *word, 
					   int len, int x, int y, int w, int h));

static void init_font (draw_struct *draw_state, 
		       clock_struct *clock_state, char *message);

static void make_window (draw_struct *draw_state);


static void
usage (void)
{
  
  fprintf (stdout, "Usage: %s [ options ]\n", progname);
  fprintf (stdout, "Where options include:\n\
\n\
  -text                         Message to display.\n\
  -timeout                      How long to display the message before\n\
                                exiting. Default is 0, which means forever.\n\
  -clock                        Enable clock display.\n\
  -12                           Display twelve hour time (default).\n\
  -24                           Display twenty-four hour time.\n\
  -seconds                      Display seconds (default).\n\
  -noseconds                    Don't display seconds.\n\
  -countdown <date>             Display a countdown instead of a clock.\n\
                                Run `-countdown foo' to see date syntax.\n\
  -countdownquit                Quit after the countdown has reached\n\
                                zero (default).\n\
  -nocountdownquit              Keep counting after reaching zero.\n\
  -geometry <geometry>          Size and position of window.\n\
  -placement <placement name>   A screen location, one of center, top,\n\
                                bottom, right, left, top-right, top-left,\n\
                                bottom-left, bottom-right. Defaults to center.\n\
                                A specified placement will override -geometry.\n\
  -font <font>                  Name of an X font to use.\n\
  -transparent                  Make the window background be transparent,\n\
                                if possible.\n\
  -nontransparent               Don't (default).\n\
  -mirror                       Flip the text about the y axis. Useful for\n\
                                back projection.\n\
  -nomirror                     Don't (default).\n\
  -fg                           Set the forground text colour (default: black).\n\
  -bg                           Set the background colour (default: white).\n\
  -bd                           Set the border colour (default: black).\n\
  -bw                           Set the border width (default: 2).\n\
  -outline                      Set the text outline width (default: 1).\n\
  -margin                       Set the margin size (default: 0).\n\n");
}

static int
splash_ehandler (Display *display, XErrorEvent *error)
{
  fprintf (stderr, "\nX error in %s:\n", progname);
  if (XmuPrintDefaultErrorMessage (display, error, stderr))
/*    exit (-1);*/
    abort();
  else
    fprintf (stderr, " (nonfatal.)\n");
  return 0;
}

/* Code taken from xdaliclock. It really needs to be hacked to handle
   ISO style dates, but maybe some other time... */
static time_t
parse_countdown(char *countdown_s)
{
  time_t countdown = 0;
  time_t now = time((time_t *) 0);
  int h100 = (99*60*60) + (59*60) + 59;
  long l = 0;
  char c = 0;
  char mon[255];
  int day = 0;
  int hour = 0;
  int min = 0;
  int sec = 0;
  int year = 0;
  *mon = 0;

  if (1 == sscanf (countdown_s, " %lu %c", &l, &c))
    countdown = l;
  else if (6 == sscanf (countdown_s, " %s %d %d:%d:%d %d %c",
			mon, &day, &hour, &min, &sec, &year, &c))
    {
      struct tm tm;
      int imonth = -1;
      const char *months[] = {"jan", "feb", "mar", "apr", "may", "jun",
			      "jul", "aug", "sep", "oct", "nov", "dec" };

      for (l = 0; mon[l]; l++)
	if (mon[l] >= 'A' && mon[l] <= 'Z')
	  mon[l] = mon[l] + ('a'-'A');
      
      for (l = 0; l < 12; l++)
	if (!strcmp(months[l], mon))
	  {
	    imonth = l;
	    break;
	  }
      if (imonth < 0)
	{
	  fprintf (stderr, "%s: unparsable month: \"%s\"\n",
		   progname, mon);
	  goto LOSE2;
	}
      if (year > 1900) year -= 1900;
      tm.tm_sec = sec;
      tm.tm_min = min;
      tm.tm_hour = hour;
      tm.tm_mday = day;
      tm.tm_mon = imonth;
      tm.tm_year = year;
      tm.tm_wday = 0;     /* ignored */
      tm.tm_yday = 0;     /* ignored */
      tm.tm_isdst = -1;
      countdown = mktime (&tm);
      if ((long) countdown < 0L)
	goto LOSE1;
    }
  else
    {
    LOSE1:
      fprintf (stderr, "%s: unparsable date: \"%s\"\n", progname, countdown_s);
    LOSE2:
      fprintf (stderr,
	       "%s: argument to -countdown should be a time_t integer\n"
	       "\t (number of seconds past \"Jan 1 00:00:00 GMT 1970\");\n"
	       "\t Or, a string of the form \"Mmm DD HH:MM:SS YYYY\",\n"
	       "\t for example, \"Jan 1 00:00:00 2000\".\n"
	       "\t This string is interpreted in the local time zone.\n",
	       progname);
      exit (1);
    }
  
  if (countdown > now + h100 ||
      countdown < now - h100)
    {
      char *s = ctime ((time_t *) &countdown);
      char *r = strchr (s, '\n');
      if (r) *r = 0;
      fprintf (stderr,
	       "%s: countdown too distant: '%s' is more\n"
	       "            than 100 hours away (can't "
	       "display it on the clock face.)\n",
	       progname, s);
      exit (1);
    }
  return countdown;
}


/* Mirrors the contents of the window about the y axis, 
   so that the text appears to be written backwards. Handy if
   your data projector doesn't have a mode to do this when
   backprojecting.
*/

static void
mirror(draw_struct *state)
{

  int x = 0;
  int new_x = state->width;
  Pixmap pix_tmp;

  if (! state->m_text_pix)
    state->m_text_pix = XCreatePixmap (state->display,
					  state->window, 
					  state->width,
					  state->height, 1);
  
  if (! state->m_outline_pix && state->outline)
    state->m_outline_pix = XCreatePixmap (state->display,
					  state->window, 
					  state->width,
					  state->height, 1);
      
  /* Copy each pixmap into it's mirror pixmap, a vertical line at a time. */
  for (x = 0; x < state->width; x++)
    {
      new_x--;
      XCopyArea(state->display, state->text_pix, 
		state->m_text_pix, state->pixmap_draw_gc,
		x, 0, 1, state->height, new_x, 0);
      if (state->outline) 
	XCopyArea(state->display, state->outline_pix, 
		  state->m_outline_pix, state->pixmap_draw_gc,
		  x, 0, 1, state->height, new_x, 0);
      
    }

  /* Swap the pixmap with the reverse pixmap */

  pix_tmp = state->text_pix;
  state->text_pix = state->m_text_pix;
  state->m_text_pix = pix_tmp;
  pix_tmp = state->outline_pix;
  state->outline_pix = state->m_outline_pix;
  state->m_outline_pix = pix_tmp;
  pix_tmp = 0;
  
}

static char *
set_clock_string(clock_struct *clock_state)
{
  char *str;
  struct tm *tm;
  
  clock_state->last_time = time ((time_t *) 0);
  tm = localtime (&clock_state->last_time);

  if (clock_state->countdown)
    {
      int delta = clock_state->countdown - clock_state->last_time;
      if (delta < 0)
	{
	if (clock_state->countdown_quit_p)
	  exit(0);	
	else	  
	  delta = -delta;
	}
      tm->tm_sec = delta % 60;
      tm->tm_min = (delta / 60) % 60;
      tm->tm_hour = (delta / (60 * 60)) % 100;
    }
  
  if (clock_state->clock_mode == 12  && tm->tm_hour > 12) tm->tm_hour -= 12;
  if (clock_state->clock_mode == 12 && tm->tm_hour == 0) tm->tm_hour = 12;

  if (clock_state->do_seconds)
    { 
      str = (char *) malloc (sizeof(char) * 8);
      sprintf(str, "%2d:%.2d:%.2d", tm->tm_hour, tm->tm_min, tm->tm_sec);
    }
  else
    {
      str = (char *) malloc (sizeof(char) * 5);
      sprintf(str, "%2d:%.2d", tm->tm_hour, tm->tm_min);
    }
  return str;
}


static void
count_word (draw_struct *state, const char *word, int len,
            int x, int y, int w, int h)
{
  if (x + w > state->max_x)
    state->max_x = x + w;
  if (y + h > state->max_y)
    state->max_y = y + h;
}


static void
draw_word (draw_struct *state, const char *word, int len,
           int x, int y, int w, int h)
{
  int xshift, yshift;
  x += state->outline;

  if (state->outline)
    {
      for (xshift = -1; xshift < 2; xshift += 2)
	{       
	  for (yshift = -1; yshift < 2; yshift += 2)
	    {		  
	      XDrawString(state->display, state->outline_pix,
			  state->pixmap_draw_gc,
			  state->xoff + x + xshift * state->outline,
			  state->yoff + y + state->font->ascent + 
			  yshift * state->outline,
			  word, len);
	    }
	}
       XDrawString (state->display, state->outline_pix, state->pixmap_draw_gc,
	       state->xoff + x,
	       state->yoff + y + state->font->ascent,
	       (char *) word, len);    
    }
  
  XDrawString (state->display, state->text_pix, state->pixmap_draw_gc,
	       state->xoff + x,
	       state->yoff + y + state->font->ascent,
	       (char *) word, len);
}


/* Loads the font, checks that it has all the numbers, if
   we're in time mode, and works out the size of the text area.
*/

static void
init_font (draw_struct *draw_state, 
	   clock_struct *clock_state, char *message)
{
  char *fn;
  int i, max_width;
  XCharStruct *ch;
  XFontStruct *font;
  XWindowAttributes xgwa;

  XGetWindowAttributes (draw_state->display, draw_state->parent, &xgwa);

  fn = get_string_resource ("font", "Font");  
  font = XLoadQueryFont (draw_state->display, fn);
  if (!font)
    {
      fprintf (stderr, "%s: couldn't load font \"%s\"\n", progname, fn);
      font = XLoadQueryFont (draw_state->display, "fixed");
      if (!font) exit (1);
    }

  draw_state->font = font;

  if (clock_state->do_clock) 
    {
    if (font->min_char_or_byte2 > '0' || 
	font->max_char_or_byte2 < '9')
      {
	fprintf (stderr, "%s: font %s doesn't contain characters '0'-'9'.\n",
		 progname, fn);
	exit (-1);
      }
    else if (font->min_char_or_byte2 > ':' || 
	     font->max_char_or_byte2 < ':')
      {
	fprintf (stderr, "%s: font %s doesn't contain ':'.\n",
		 progname, fn);
	exit (-1);
      }

    max_width = font->min_bounds.lbearing + font->min_bounds.rbearing;
    for (i = '0'; i <= '9'; i++)
      {
	ch = (font->per_char
	      ? &font->per_char [i - font->min_char_or_byte2]
	      : &font->min_bounds);
	max_width = MAX (max_width, ch->width);
	if (ch->width == 0)
	  {
	    fprintf (stderr,
		     "%s: char '%c' has width %d in font %s\n",
		     progname, i, ch->width, fn);
	    exit (-1);
	  }
      }
     
    /* Yucky - check that ':' is present */
    ch = (font->per_char
	  ? &font->per_char [':' - font->min_char_or_byte2]
	  : &font->min_bounds);
    if (ch->width == 0)
      {
	fprintf (stderr,
		 "%s: char '%c' has width %d in font %s\n",
		 progname, ':', ch->width, fn);
	exit (-1);
      }


    draw_state->max_x = (clock_state->do_seconds ? 
			 max_width * 6 + CHAR_W(':') * 2
			 : max_width * 4 + CHAR_W(':')); 
    draw_state->max_y = font->descent + font->ascent + 1;

    }  
  else
    /* Set up draw_state->max_x, draw_state->max_y, the text block size */
    iterate_words (draw_state, message, 
		   xgwa.width - (draw_state->margin * 2), count_word);

  
  /* Add any outline size */
  draw_state->max_x += (2 * draw_state->outline);
  draw_state->max_y += (2 * draw_state->outline);
  

}


static void
iterate_words (draw_struct *state,
	       const unsigned char *string,	       
               int max_width,
               void (*mapper) (draw_struct *state,
                               const char *word, int len,
                               int x, int y, int w, int h)
               )
{  
  int x = 0, y = 0;
  XFontStruct *font = state->font;
  int line_height = font->ascent + font->descent + 1;

  while (*string)
    {
      const unsigned char *start;
      int word_w = 0;

      while (*string == ' ' || *string == '\t')		  /* skip hspace */
        {
          x += CHAR_W (' ');
          string++;
        }

      start = string;
      while (*string != 0 &&				  /* skip non-hspace */
             *string != ' ' &&
             *string != '\t' &&
             *string != '\n')
        {
          word_w += CHAR_W (*string);
          string++;
        }

      if (max_width > 0 && (x + word_w) > max_width)
        {
          x = 0;
          y += line_height;
        }

      if (start != string)
        mapper (state, start, string-start, x, y, word_w, line_height);

      if (*string == '\n' || *string == 0)
        {
          x = 0;
          y += line_height;
          if (*string) string++;
          while (*string == ' ' || *string == '\t')   /* swallow whitespace */
            string++;
        }
      else
        {
          x += word_w;
        }
    }
}
              
static void
initialize_draw_state (draw_struct *draw_state,
		       clock_struct *clock_state,
		       char *message)
{

  int x, y, w, h;
  char *geometry, *placement;
  int xoff=0, yoff=0, georeturn=0;
  uint width=0, height=0;

  XWindowAttributes xgwa;
  XGCValues gcv;

  XGetWindowAttributes (draw_state->display, draw_state->parent, &xgwa);

  init_font(draw_state, clock_state, message);

  placement = get_string_resource ("placement", "Placement");
  if (!placement)
    {
      /* overrides geometry */
      geometry = get_string_resource ("geometry", "Geometry");
      if (geometry)
        georeturn = XParseGeometry(geometry, &xoff, &yoff, &width, &height);
    }

  w = (georeturn & WidthValue) ? width 
    : draw_state->max_x + (draw_state->margin * 2);
  h = (georeturn & HeightValue) ? height 
    : draw_state->max_y + (draw_state->margin * 2);
  if (georeturn & XValue)
    x = (georeturn & XNegative) ? (xgwa.width - xoff - w) : xoff;
  else
    /* Default to center */
    x = (xgwa.width  - w) / 2;

  if (georeturn & YValue)
    y = (georeturn & YNegative) ? (xgwa.height - yoff - h) : yoff;
  else
    /* Default to center */
    y = (xgwa.height - h) / 2;

  if (x < 0) x = 0;
  if (y < 0) y = 0;

  if (placement)
    {
      if (strcasecmp(placement, "center") == 0)
        ;  /* do nothing, since that's the default above anyway */
      else if (strcasecmp(placement, "top") == 0)
        y = 0;
      else if (strcasecmp(placement, "bottom") == 0)
        y = (xgwa.height - h);
      else if (strcasecmp(placement, "left") == 0)
        x = 0;
      else if (strcasecmp(placement, "right") == 0)
        x = (xgwa.width  - w);
      else if (strcasecmp(placement, "top-left") == 0)
	{
	  y = 0; 
	  x = 0;
	}
      else if (strcasecmp(placement, "top-right") == 0)
	{
	  y = 0; 
	  x = (xgwa.width - w); 
	}
      else if (strcasecmp(placement, "bottom-left") == 0)
	{
	  y = (xgwa.height - h); 
	  x = 0;
	}
      else if (strcasecmp(placement, "bottom-right") == 0)
	{
	  y = (xgwa.height - h); 
	  x = (xgwa.width  - w);
	}
      else
	; /* do center, which is the default */
    }

  /* Try to center the text block inside the final window dimensions */
  draw_state->xoff = (w - draw_state->max_x) / 2;
  draw_state->yoff = (h - draw_state->max_y) / 2;
  if (draw_state->xoff < 0) draw_state->xoff = 0;
  if (draw_state->yoff < 0) draw_state->yoff = 0;

  draw_state->x = x;
  draw_state->y = y;
  draw_state->width = w;
  draw_state->height = h;

  /* Resize and position the window, and create the pixmap */
  XMoveResizeWindow (draw_state->display, 
		     draw_state->window, x, y, w, h);
  draw_state->text_pix = XCreatePixmap (draw_state->display, 
				      draw_state->window, w, h, 1);
  if (draw_state->outline)
    draw_state->outline_pix = XCreatePixmap (draw_state->display, 
					     draw_state->window, w, h, 1);
  
  /* Set up the GCs */  
  gcv.font = draw_state->font->fid;
  gcv.graphics_exposures = False;

  /* This GC fills in just the body of the outline with the
     background colour, as the outline is in the foreground color.
  */
  if (draw_state->outline)
    {
      gcv.function = GXandReverse;
      gcv.foreground = 1;
      gcv.background = 0;
      draw_state->outline_draw_gc = XCreateGC (draw_state->display, 
					      draw_state->text_pix, 
					      GCFont | GCForeground | 
					      GCBackground |
					      GCFunction |
					      GCGraphicsExposures , &gcv);
    }

  gcv.foreground = draw_state->fg;
  gcv.background = draw_state->bg; 
  gcv.function = GXcopy;
  draw_state->window_draw_gc = XCreateGC (draw_state->display,
					  draw_state->window,
					  GCFont | GCForeground | 
					  GCBackground |
					  GCFunction |
					  GCGraphicsExposures, &gcv);
  gcv.foreground = draw_state->bg;
  gcv.background = draw_state->fg;
  draw_state->window_erase_gc = XCreateGC (draw_state->display, 
					   draw_state->window, 
					   GCFont | GCForeground | 
					   GCBackground |
					   GCFunction | 
					   GCGraphicsExposures, &gcv);

  gcv.foreground = 1;
  gcv.background = 0;
  draw_state->pixmap_draw_gc = XCreateGC (draw_state->display, 
					  draw_state->text_pix, 
					  GCFont | GCForeground | 
					  GCBackground |
					  GCFunction |
					  GCGraphicsExposures, &gcv);
  gcv.foreground = 0;
  gcv.background = 1;
  draw_state->pixmap_erase_gc = XCreateGC (draw_state->display, 
					   draw_state->text_pix, 
					   GCFont | GCForeground | 
					   GCBackground |
					   GCFunction |
					   GCGraphicsExposures , &gcv);

}


static void
draw(draw_struct *state, char *message)
{
  if (state->changed_p)
    {
      XFillRectangle (state->display, state->text_pix,
		      state->pixmap_erase_gc, 0, 0,
		      state->width, state->height);
      if (state->outline)
	XFillRectangle (state->display, state->outline_pix,
			state->pixmap_erase_gc, 0, 0,
			state->width, state->height);
      
      iterate_words (state, message, state->width - (state->margin * 2), draw_word);
      if (state->mirror_p)
	mirror(state);
    }

#if HAVE_SHAPE
  if (state->do_shape && state->changed_p)
  {      
    if (state->outline)	    
      XShapeCombineMask (state->display, state->window, ShapeBounding,
			 0, 0, state->outline_pix, ShapeSet);	    
    else
      XShapeCombineMask (state->display, state->window, ShapeBounding,
			 0, 0, state->text_pix, ShapeSet);
  }
#endif
  {
    XMapRaised (state->display, state->window);
    if (state->changed_p && state->outline)
      {
	XCopyArea (state->display, state->outline_pix,
		    state->text_pix, state->outline_draw_gc,
		    0, 0, state->width, state->height,
		    0, 0);
      }
    
    XCopyPlane (state->display, state->text_pix,
		state->window, state->window_draw_gc,
		0, 0, state->width, state->height,
		0, 0, 1);    
  }
  state->changed_p = 0;
}

static void
alarm_handler (int sig)
{
  exit (0);
}


static int
check_events (draw_struct *state)
{
  XEvent event;
  int redraw_p = 0;

  while (XPending(state->display))
    {
      XNextEvent (state->display, &event);
      switch (event.xany.type)
	{
	case KeyPress:
	case ButtonRelease:  /* wait for the "up" transition */
	  exit (0);
	  break;
	case ClientMessage:
	  if (event.xclient.message_type == XA_WM_PROTOCOLS &&
              event.xclient.data.l[0] == XA_WM_DELETE_WINDOW)
	    exit (0);
	  break;
	case Expose:	  
	  redraw_p = 1;
	  break;
	case VisibilityNotify:
	  if (event.xvisibility.state == VisibilityFullyObscured) 
	    {
	      XMapRaised (state->display, state->window);
	      redraw_p = 1;
	    }
	  break;
	}      
    }
  return redraw_p;
}

static void
make_window (draw_struct *draw_state)
{  
  XSetWindowAttributes attrs;  
  unsigned long attrmask;
  Colormap cmap;
  Visual *visual = 0;
     
  /* Transparency stuff */
  if (draw_state->do_shape)
    {
      unsigned long overlay_transparent_pixel = 0;     
      Visual *v = get_overlay_visual(DefaultScreenOfDisplay(draw_state->display),
				     &overlay_transparent_pixel);
      if (v)
	{
	  draw_state->do_shape = 0;
	  draw_state->do_overlay = 1;
	  visual = v;
	}
    }

  if (!draw_state->do_overlay)    
    visual = CopyFromParent;
  
  cmap = DefaultColormap(draw_state->display, DefaultScreen(draw_state->display));
  
  draw_state->fg = get_pixel_resource("foreground","Foreground",
				      draw_state->display, cmap);
  draw_state->bg = get_pixel_resource("background", "Background",
				      draw_state->display, cmap);
  draw_state->bd = get_pixel_resource("borderColor", "Bordercolor",
				      draw_state->display, cmap);

#ifdef HAVE_SHAPE
  {
    int shape_event_base, shape_error_base;
    if ((draw_state->do_shape || draw_state->do_overlay) &&
        XShapeQueryExtension (draw_state->display, 
			      &shape_event_base, &shape_error_base))
      ;
    else if (draw_state->do_shape)
      {
        fprintf (stderr, "%s: no shape extension on this display\n", progname);
        draw_state->do_shape = 0;
      }
  }
#endif /* HAVE_SHAPE */

  attrmask = (CWOverrideRedirect | CWEventMask | CWBorderPixel | CWBackPixel |
	      CWBackingStore | CWSaveUnder | CWBackingPixel);
  attrs.override_redirect = True;
  attrs.event_mask = (KeyPressMask | KeyReleaseMask |
		      ButtonPressMask | ButtonReleaseMask |
		      ExposureMask | VisibilityChangeMask |
                      StructureNotifyMask);

  attrs.save_under = True;
  attrs.backing_store = NotUseful;
  if (draw_state->do_shape) 
    {
      attrs.border_pixel = draw_state->bd;
      attrs.background_pixel = draw_state->fg;
      attrs.backing_pixel = draw_state->fg;
    }
  else
    {
      attrs.border_pixel = draw_state->bd;
      attrs.background_pixel = draw_state->bg;
      attrs.backing_pixel = draw_state->bg;
    }
  draw_state->window = XCreateWindow (draw_state->display, draw_state->parent,
				      draw_state->x, draw_state->y, 
				      draw_state->width, draw_state->height,
				      draw_state->border,
				      CopyFromParent,
				      InputOutput,
				      visual,
				      attrmask, &attrs);
  
  XSelectInput (draw_state->display, draw_state->parent, 
		(VisibilityChangeMask | StructureNotifyMask));
}

int
main (int argc, char **argv)
{
  XtAppContext app;
  Display *display;
  Widget toplevel;

  draw_struct *draw_state = (draw_struct *) calloc (1, sizeof(*draw_state));
  clock_struct *clock_state = (clock_struct *) calloc (1, sizeof(*clock_state));
  
  char *countdown_s, *message;
  int timeout = 0;
  int redraw_p = 0;
    
  progname = argv[0];
  toplevel = XtAppInitialize (&app, progclass, options, countof(options),
                              &argc, argv, defaults, 0, 0);
  display = XtDisplay (toplevel);
  db = XtDatabase (display);

  XtGetApplicationNameAndClass (display, &progname, &progclass);
  XSetErrorHandler (splash_ehandler);

  if (argc > 1)
    {
      usage();
      exit (1);
    }
  
  /* Init the clock state from resources*/
  clock_state->do_clock = get_boolean_resource("useclock", "Useclock");
  clock_state->clock_mode = get_integer_resource ("clockmode", "Clockmode");
  clock_state->do_seconds = get_boolean_resource ("seconds", "Seconds");
  clock_state->countdown_quit_p = get_boolean_resource ("countdownquit",
						       "Countdownquit");

  countdown_s = get_string_resource ("countdown", "Countdown");
  if (countdown_s && *countdown_s) 
    clock_state->countdown = parse_countdown(countdown_s);
  else
    clock_state->countdown = 0;
  if (countdown_s) free(countdown_s);

  if (clock_state->countdown)
    {
      clock_state->do_clock = 1;
      clock_state->clock_mode = 24;
    }

  if (clock_state->do_clock)
    clock_state->last_time = time ((time_t *) 0);
  else
    clock_state->last_time = 0;
  

  /* Init the draw state from resources*/
  draw_state->display = display;
  draw_state->do_shape = get_boolean_resource ("transparent", "Transparent");

  /* We defer colours until making the window, as we need a colormap to work
     with. */

  draw_state->margin = get_integer_resource ("margin", "Margin");
  if (draw_state->margin < 0)
    {
      fprintf(stderr, "%s: The margin must be zero or greater.\n", progname);
      exit(1);
    }
  draw_state->border = get_integer_resource ("borderWidth", "BorderWidth");
  if (draw_state->border < 0)
    {
      fprintf(stderr, "%s: The border must be zero or greater.\n", progname);
      exit(1);
    }
  draw_state->outline = get_integer_resource ("outlineWidth", "OutlineWidth");
  if (draw_state->outline < 0)
    {
      fprintf(stderr, "%s: The outline must be zero or greater.\n", progname);
      exit(1);
    } 

  draw_state->mirror_p = get_boolean_resource ("mirror", "Mirror");
  draw_state->changed_p = 1;

  draw_state->xoff = 0;
  draw_state->yoff = 0;
  draw_state->max_x = 0;
  draw_state->max_y = 0;
  draw_state->x = 0;
  draw_state->y = 0;
  draw_state->width = 1;
  draw_state->height = 1;


  XA_WM_PROTOCOLS = XInternAtom (display, "WM_PROTOCOLS", False);
  XA_WM_DELETE_WINDOW = XInternAtom (display, "WM_DELETE_WINDOW", False);

  draw_state->parent = RootWindow (display, DefaultScreen (display));
  /* Make our child window */
  make_window (draw_state);

  if (clock_state->do_clock)
    message = set_clock_string(clock_state);    
  else
    message = get_string_resource ("text", "Text");

  /* Set up the drawables - window size, pixmaps, font */
  initialize_draw_state(draw_state, clock_state, message);

  draw (draw_state, message);

  timeout = get_integer_resource ("timeout", "Timeout");
  if (timeout > 0)
    {
      signal (SIGALRM, alarm_handler);
      alarm (timeout);
    }

  while(1)
    {
      /* Pilfered from xdaliclock. This keeps similarly coded clocks
	 in lock-step with one another.
      */
      long now = 0;
      TICK_SLEEP ();
      now = time ((time_t *) 0);
      if (clock_state->do_clock && clock_state->last_time != now) 
	{
	  if (message) free(message);    
	  message = set_clock_string(clock_state);
	  draw_state->changed_p = 1;
	  redraw_p = 1;
	}
      if (check_events(draw_state))
	redraw_p = 1;
	
      if (redraw_p)
	{
	  draw(draw_state, message);
	  redraw_p = 0;
	}
    }

  return 0;
}
